// WebServer module
// Partly based on: // https://gist.github.com/me-no-dev/d34fba51a8f059ac559bf62002e61aa3
//
#include <WiFi.h>
#include <ESPAsyncWebServer.h>  // Local WebServer used to server the configuration portal
#include <Update.h>
#include <SPIFFS.h>
//#include "ArduinoJson.h"
#include "SD_MMC.h"            // SD Card ESP32

#include "Main.h"
#include "WebServer.h"
#include "cam.h"
#include "hw.h"
#include "cfg.h"
#include "session.h"

char *skip_spc(char *p)
{
	while(*p)
	{
		if(*p!=' ' && *p!='\t' && *p!='\r' && *p!='\n') return p;
		p++;
	}
	return p;
}

String GetCookieName(char *cookie, char *name)
{
	char tmp[1024];
	char *p , *p_start,*p_val;

	if(cookie && name)
	{
		strncpy(tmp,cookie,sizeof(tmp)-1);
		tmp[sizeof(tmp)-1]='\0';
		p=tmp;
		p_start=skip_spc(p);
		p_val=p_start;

		while(*p)
		{
			if(*p==';')
			{
				*p='\0';
				if(strcmp(p_start,name)==0)
				{
					return String(p_val);
				}
				p++;
				p=skip_spc(p);
				p_start=p;
			}
			else if(*p=='=')
			{
				*p='\0';
				p++;
				p=skip_spc(p);
				p_val=p;
			}
			else
			{
				p++;
			}
		}
		if(strcmp(p_start,name)==0)
		{
			return String(p_val);
		}
	}

	return "";
}

String makePage(String title, String contents) {
	String s = F("\n<!DOCTYPE html><html><head>");

	s += F("<meta name=\"viewport\" content=\"width=device-width,user-scalable=0\">");
	s += F("<title>");
	s += title;
	s += F("</title></head><body>\n");
	s += contents;
	s += F("\n</body></html>\n");
	return s;
}

int8_t GetSidCookie(AsyncWebServerRequest *request)
{
	if(request->hasHeader("Cookie"))
	{
		AsyncWebHeader* h = request->getHeader("Cookie");

		String val=GetCookieName((char *)(h->value().c_str()), (char *)"_ESID");
		if(val!="")
		{
			int8_t n=session_search((char *)(val.c_str()));
			if(n>=0)
			{
				Log.printf("Cookie _ESID Found: [%s]\r\n",val.c_str());
			}
			else
			{
				Log.printf("Cookie _ESID Found but Wrong: [%s]\r\n", val.c_str());
			}
			return n;
		}
		Log.printf("Cookie _ESID not Found: [%s]\r\n", h->value().c_str());
	}
	return -2;
}

bool TestSidCookie(AsyncWebServerRequest *request)
{
	int8_t n=GetSidCookie(request);

	if(n>=0)
	{
		return true;
	}
	return false;
}

bool _webAuth_auth(AsyncWebServerRequest *request, bool en_auth_var) {
	/*
	// Print header for debugging
	Log.printf("url:%s Headers: %s:\n", request->url().c_str(), String(request->headers()).c_str());
	for (uint i = 0; i < request->headers(); i++)
	{
		Log.printf(" %i  %s=%s\n", i, request->headerName(i).c_str(), request->header(i).c_str());
	}
	//
	*/
	int headers = request->headers();
	int i;
	for(i=0;i<headers;i++){
		AsyncWebHeader* h = request->getHeader(i);
		Log.printf("HEADER[%s]: [%s]\n\r", h->name().c_str(), h->value().c_str());
	}

	if(en_auth_var)
	{
		// only for get from file or not
		if(request->hasParam("auth", false, true))
		{
			AsyncWebParameter* p = request->getParam("auth", false, true);
			int8_t n=session_search((char *)(p->value().c_str()));
			Log.printf("Param file auth Found: [%d]\r\n",n);
			if(n>=0)
			{
				return true;
			}
		}
		else if(request->hasParam("auth", false, false))
		{
			AsyncWebParameter* p = request->getParam("auth", false, false);
			int8_t n=session_search((char *)(p->value().c_str()));
			Log.printf("Param auth Found: [%d]\r\n",n);
			if(n>=0)
			{
				return true;
			}
		}
	}

	if(request->hasHeader("authorization"))
	{
		char *p, *p1;
		AsyncWebHeader* h = request->getHeader("authorization");
		p=(char *)(h->value().c_str());
		Log.printf("AUTH Found: [%s]\r\n",p);
		if((p1=strstr(p,"Bearer")))
		{
			p1+=6;
			p1=skip_spc(p1);
			int8_t n=session_search(p1);
			Log.printf("Param Bearer Found: [%s] %d\r\n",p1,n);
			if(n>=0)
			{
				return true;
			}
		}
	}

	if(TestSidCookie(request))
	{
		return true;
	}
	return false;
}

bool _webAuth(AsyncWebServerRequest *request, bool en_auth_var) {
	bool res = _webAuth_auth(request, en_auth_var);
	if(!res)
	{
		if (!request->authenticate(cfg_global.un.c_str(), cfg_global.up.c_str(), NULL, false)) 
		{
		    request->requestAuthentication(NULL, false);  // force basic auth
		    return false;
		}
		return true;
	}
	return res;
}

bool _webAuth(AsyncWebServerRequest *request) {
	return _webAuth(request, false);
}

bool _webAuth_var(AsyncWebServerRequest *request) {
	return _webAuth(request, true);
}

/**************************************************************************
 *  Webfunction proof AuthCookie
 * ************************************************************************/
void AuthGet(AsyncWebServerRequest *request) {
	Log.printf("Auth Get %d\r\n",request->params());

	bool res = _webAuth_auth(request, true);

	// test for Cookie
	if(res)
	{
		request->send(200, F("application/json"), F("{\"success\":true}"));
		return;
	}
	request->send(200, F("application/json"), F("{\"success\":false}"));
}

/**************************************************************************
 *  Webfunction Login an set  AuthCookie
 * ************************************************************************/
void AuthPost(AsyncWebServerRequest *request) {
	Log.printf("Auth Post %d\r\n",request->params());
session_print();

	if(request->hasParam("name", true))
	{
		AsyncWebParameter* n = request->getParam("name", true);
		if(request->hasParam("pw", true))
		{
			AsyncWebParameter* p = request->getParam("pw", true);
			if(p->value()==cfg_global.up && n->value()==cfg_global.un)
			{
				// test if valid cookie is set
				if(!TestSidCookie(request))
				{
					// gen sid
					char tmp1[128];
					uint8_t nr= session_gen();
					sprintf(tmp1,"{\"success\":true,\"_ESID\":\"%s\"}",Sessions[nr].sid);
					AsyncWebServerResponse *response = request->beginResponse(200,F("application/json"),tmp1);
					session_print();
					request->send(response);
					return;
				}
				// gen session or use old one and send it 
session_print();
				request->send(200, F("application/json"), F("{\"success\":true}"));
				return;
			}
		}
	}
session_print();
	request->send(200, F("application/json"), F("{\"success\":false}"));
}

/**************************************************************************
 *  Webfunction Login an set  AuthCookie
 * ************************************************************************/
void AuthDel(AsyncWebServerRequest *request) {
	Log.printf("Auth Del %d\r\n",request->params());
session_print();
	int8_t n=GetSidCookie(request);
	if(n>=0)
	{
		session_clear_n(n);
		session_write();

		AsyncWebServerResponse *response = request->beginResponse(200,F("application/json"),F("{\"success\":true}"));
//		response->addHeader(F("Set-Cookie"),F("_ESID=; path=/; expires=Thu, 01 Jan 1970 00:00:00 GMT"));
		session_print();
		request->send(response);
		return;
	}
session_print();
	request->send(200, F("application/json"), F("{\"success\":false}"));
}

void h_wificlear(AsyncWebServerRequest *request) {
	if (!_webAuth(request))
		return;
	Log.printf("wificlear %d\r\n",request->params());
	WiFi.disconnect(true,true);
	reboot = true;
	rebootdelay = millis();

	request->send(200, F("application/json"), F("{\"success\":true}"));
}

void h_reboot(AsyncWebServerRequest *request) {
	if (!_webAuth(request))
		return;
	Log.printf("reboot %d\r\n",request->params());
	reboot = true;
	rebootdelay = millis();

	request->send(200, F("application/json"), F("{\"success\":true}"));
}

void h_logout(AsyncWebServerRequest *request) {
	if (!_webAuth(request))
		return;
	Log.printf("logout %d\r\n",request->params());
	String s = F("Logout done.");
	request->send(401, "text/html", makePage(esp_name, s));
}


/**************************************************************************
 *  Webfunction to play sound
 * ************************************************************************/
void PlaySnd(AsyncWebServerRequest *request) {
	int8_t nr=0;

	if (!_webAuth(request))
		return;
	Log.printf("SND param %d\r\n",request->params());

	if (request->hasParam("p",true) && IS_SOUND_BUSY)
	{
		String p=request->getParam("p",true)->value();
		nr = cfg_SearchSnd((char *)(p.c_str()));
		if(nr>=0)
		{
			Log.printf("SND %d\r\n",nr);
			work_snd_busy=true;
			if(cfg_global.snd_on)
			{
				hw_play_snd(atoi(cfg_global.sound[nr].val.c_str()));
			}
			if(cfg_global.sound[nr].open==true)
			{
				work_do_open_after_snd=true;
			}
			request->send(200, F("application/json"), F("{\"success\":true}"));
			return;
		}
	}

	Log.printf("SND no nr\r\n");
	request->send(200, F("application/json"), F("{\"success\":false}"));
}

/**************************************************************************
 *  Webfunction to open door
 * ************************************************************************/
void DoorOpen(AsyncWebServerRequest *request) {
	if (!_webAuth(request))
		return;
	Log.printf("Door Open %d\r\n",request->params());
	if(IS_OPEN_BUSY)
	{
		request->send(200, F("application/json"), F("{\"success\":false}"));
		return;
	}
	work_do_open=true;
	request->send(200, F("application/json"), F("{\"success\":true}"));
}

/**************************************************************************
 *  Get the Config and store on SPIFFS. Read the new Values
 * ************************************************************************/
void CfgPost(AsyncWebServerRequest *request) {
	uint16_t nr=0;

	// only posible with basic auth
	if (!request->authenticate(cfg_global.un.c_str(), cfg_global.up.c_str(), NULL, false)) 
	{
	    request->requestAuthentication(NULL, false);  // force basic auth
	    return;
	}
//	if (!_webAuth(request))
//		return;
	nr=request->params();
	Log.printf("CfgPost param %d\r\n",nr);
	if(nr)
	{
		File file = SPIFFS.open("/cfg.json", "w");
		if (!file || file.isDirectory()) {
			Log.println(F("Fail to open cfg.json"));
		}
		else
		{
			AsyncWebParameter* p=request->getParam(0);
			String pt=p->value();
			if (!file.print(pt)) {
				file.close();
				Log.println(F("Fail to save cfg.json"));
			}
			else
			{
				file.close();
				cfg_parse((char *)pt.c_str());
				hw_cfg_load();
				request->send(200, F("application/json"), F("{\"success\":true}"));
				return;
			}
		}
	}

	request->send(200, F("application/json"), F("{\"success\":false}"));
}

/**************************************************************************
 *  Webfunction to get the Config
 * ************************************************************************/
void CfgGet(AsyncWebServerRequest *request) {

	// only posible with basic auth else only send subset of cfg
	if (!request->authenticate(cfg_global.un.c_str(), cfg_global.up.c_str(), NULL, false)) 
	{
		char jbuf[2000];
		uint16_t len=0;
		uint8_t i;
		bool first=true;

		if (!_webAuth(request))
			return;
		len += sprintf(jbuf+len,"{\"opener_on\":%s,\"snd_on\":%s,\"sound\":[",cfg_global.opener_on ? "true" : "false",cfg_global.snd_on ? "true" : "false");
		for(i=0; i<CFG_SND_MAX;i++)
		{
			if(cfg_global.sound[i].name[0]!='\0')
			{
				len += sprintf(jbuf+len,"%s{\"name\":\"%s\",\"ring\":%s,\"open\":%s,\"val\":\"%s\",\"description\":\"%s\"}",
					first ? "" : ",",
					cfg_global.sound[i].name.c_str(),
					cfg_global.sound[i].ring ? "true" : "false",
					cfg_global.sound[i].open ? "true" : "false",
					cfg_global.sound[i].val.c_str(),
					cfg_global.sound[i].description.c_str()
				);
				first=false;
			}
		}
		len += sprintf(jbuf+len,"]}");

		request->send(200, F("application/json"),jbuf);
		return;
	}

//	if (!_webAuth(request))
//		return;
	Log.printf("CfgGet\r\n");
	request->send(SPIFFS, "/cfg.json", String(), true);
}

/**************************************************************************
 *  Webfunction to get the last Image on SD. Only valid if Take Picture after reboot
 * ************************************************************************/
void LastGet(AsyncWebServerRequest *request) {
	char r[64];
	if (!_webAuth(request))
		return;
	Log.printf("LastImg\r\n");
	sprintf(r,"{\"success\":true,\"last\":\"%s\"}",LastImg);
	request->send(200, F("application/json"), r);
}

/**************************************************************************
 *  Webfunction to make a new Picture an Store it to SD
 * ************************************************************************/
void PhotoGet(AsyncWebServerRequest *request) {

	if (!_webAuth(request))
		return;
	Log.printf("PhotoGet\r\n");
	MakePhoto();
	request->send(200, F("application/json"), F("{\"success\":true}"));
}

/**************************************************************************
 *  Webfunction to delete Sessions
 * ************************************************************************/
void SessionDel(AsyncWebServerRequest *request) {
	if (!_webAuth(request))
		return;
	Log.println("SessionDelete");
	session_del();
	request->send(200, F("application/json"), F("{\"success\":true}"));
	return;
}

/**************************************************************************
 *  Webfunction to delete one Pictore
 * ************************************************************************/
void ImgDel(AsyncWebServerRequest *request) {
	if (!_webAuth(request))
		return;
	Log.println("DelPhoto");
	if (request->hasParam("p",true))
	{
		String p=request->getParam("p",true)->value();
		if(strcmp(LastImg,p.c_str())==0)
		{
			// delete LastImg
			*LastImg='\0';
		}
		p="/"+p;
		Log.println(p);
		SD_MMC.remove(p);
		request->send(200, F("application/json"), F("{\"success\":true}"));
		return;
	}
	request->send(200, F("application/json"), F("{\"success\":false}"));
}

/**************************************************************************
 *  Webfunction to get a Dir-Listing of stored Pictures on SD
 * ************************************************************************/
void DirGet(AsyncWebServerRequest *request) {

	static bool is_start;
	static bool is_end;
	static File root;
	static File file;
	static uint16_t pos;

	if (!_webAuth(request))
		return;

	Log.printf("DirGet\r\n");

	if(root)
	{
		root.close();
	}

	root = SD_MMC.open("/");
	if (!root || !root.isDirectory()) {
		Log.println(F("Fail to open /img/"));
		request->send(200, F("application/json"), F("{\"success\":false}"));
		return;
	}

	root.rewindDirectory();
	file = root.openNextFile();
	is_start=true;
	is_end=false;

	AsyncWebServerResponse *response = request->beginChunkedResponse(F("application/json"), [](uint8_t *buffer, size_t maxLen, size_t index) -> size_t {
		//Write up to "maxLen" bytes into "buffer" and return the amount written.
		//index equals the amount of bytes that have been already sent
		//You will be asked for more data until 0 is returned
		//Keep in mind that you can not delay or yield waiting for more data!
		pos=0;
		if(is_start)
		{
			pos+=sprintf((char *)(buffer+pos),"{\"elems\":[");
		}
		while (file) { // Now get all the filenames, file types and sizes
			if(file.isDirectory())
			{
				file = root.openNextFile();
			}
			else
			{
				if(!is_start)
				{
					*(buffer+pos)=',';
					pos++;
				}
				is_start=false;
				pos+=sprintf((char *)(buffer+pos),"\"%s\"",file.name()+1);
				file = root.openNextFile();
				if(pos > maxLen-1000)
				{
					break;
				}
			}	
		}
		if(!file && !is_end)
		{
			pos+=sprintf((char *)(buffer+pos),"]}\n");
			is_start=false;
			is_end=true;
		}
		Log.println("Sum: "+String(index)+" "+String(maxLen)+" "+String(pos));
		if(!pos)
		{
			root.close();
		}
		return pos;
	});
	request->send(response);

}

// Define all specific webpages to service by the ESP
void WebServerInit(AsyncWebServer *webserver) {

	// Define all weboptions with TEMPLATE variables or specific functions
	webserver->on("/wificlear", HTTP_GET, h_wificlear);
	webserver->on("/reboot", HTTP_GET, h_reboot);
	webserver->on("/logout", HTTP_GET, h_logout);

	// Cam links
	webserver->on("/capture", HTTP_GET, sendJpg);
	webserver->on("/stream", HTTP_GET, streamJpg);

	webserver->on("/snd", HTTP_POST, PlaySnd);
	webserver->on("/cfg", HTTP_POST, CfgPost);
	webserver->on("/open", HTTP_POST, DoorOpen);
	webserver->on("/cfg", HTTP_GET, CfgGet);
	webserver->on("/dir", HTTP_GET, DirGet);
	webserver->on("/lastimg", HTTP_GET, LastGet);
	webserver->on("/photo", HTTP_GET, PhotoGet);
	webserver->on("/photo", HTTP_DELETE, ImgDel);
	webserver->on("/session", HTTP_DELETE, SessionDel);

	webserver->on("/auth", HTTP_GET, AuthGet);
	webserver->on("/auth", HTTP_POST, AuthPost);
	webserver->on("/auth", HTTP_DELETE, AuthDel);

	webserver->on("/cfg.json", HTTP_GET, WrongPage);
	webserver->on("/cfg.json", HTTP_POST, WrongPage);
	webserver->on("/esid", HTTP_GET, WrongPage);
	webserver->on("/esid", HTTP_POST, WrongPage);

	// serving other static information from SPIFFS
	webserver->serveStatic("/", SPIFFS, "/")
		.setAuthentication(_webAuth_var);
//		.setAuthentication(cfg_global.un.c_str(), cfg_global.up.c_str());

	webserver->serveStatic("/img/", SD_MMC, "/")
		.setAuthentication(_webAuth_var);
//		.setAuthentication(cfg_global.un.c_str(), cfg_global.up.c_str());

	// process invalid requests
	webserver->onNotFound(WrongPage);
	// Start Webserver
	webserver->begin();
}

// ----------------------------------------------------------------------------------------------------------------
// ------  General Webfunctions -----------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------

void WrongPage(AsyncWebServerRequest *request) {
	//Log.print(request->url());
	//Log.println(F(" -> invalid request"));
	String s = F("<h1>");
	s += esp_name;
	s += F("</h1><p>Wrong request</p>");
	s += F("<p><a href=\"/\">Home</a></p>");
	request->send(500, "text/html", makePage(esp_name, s));
}

String urlDecode(String input) {
	String s = input;

	s.replace("%20", " ");
	s.replace("+", " ");
	s.replace("%21", "!");
	s.replace("%22", "\"");
	s.replace("%23", "#");
	s.replace("%24", "$");
	s.replace("%25", "%");
	s.replace("%26", "&");
	s.replace("%27", "\'");
	s.replace("%28", "(");
	s.replace("%29", ")");
	s.replace("%30", "*");
	s.replace("%31", "+");
	s.replace("%2C", ",");
	s.replace("%2E", ".");
	s.replace("%2F", "/");
	s.replace("%2C", ",");
	s.replace("%3A", ":");
	s.replace("%3A", ";");
	s.replace("%3C", "<");
	s.replace("%3D", "=");
	s.replace("%3E", ">");
	s.replace("%3F", "?");
	s.replace("%40", "@");
	s.replace("%5B", "[");
	s.replace("%5C", "\\");
	s.replace("%5D", "]");
	s.replace("%5E", "^");
	s.replace("%5F", "-");
	s.replace("%60", "`");

	// int str_len = s.length() + 1;
	// char char_array[str_len];
	// s.toCharArray(char_array, str_len);
	return s;
}
