#include "Arduino.h"
#include "json.h"
#include <TelnetSpy.h>

extern TelnetSpy LOG;
#define Log LOG

#define JSON_SQUOTE	0
#define JSON_QUOTE	1
#define JSON_VAL	2
#define JSON_ARRAY	3

char *jsonbuf_skip_white(char *src)
{
	while(*src)
	{
		if(*src>' ')
		{
			return src;
		}
		src++;
	}
	return src;
}

void jsonbuf_print_elem(char *token,char *val)
{
	Log.print(token);
	Log.print(F(": "));
	Log.println(val);
}

char *jsonbuf_parse(char *src, char *token, char *rdst, uint8_t rdst_len, int8_t n)
{
	uint8_t typ=0;
	char *p, *pt ,*pdst;
	bool have_token=false;
	bool do_val=false;
	char *res=NULL;
	uint8_t token_count=0;
	uint8_t array_count=0;

	pdst=rdst;

	if(pdst)
	{
		*pdst='\0';
	}

	if(!src)
	{
		return NULL;
	}

	// scan for the token
	p=src;
	pt=token;

	while(*p)
	{
		if(!have_token)
		{
			if(*p == *pt)
			{
				// same char at token
				pt++;
				if(*pt=='\0')
				{
					// token end
					p++;
					if(*p=='"' || *p=='\'')
					{
						// skip quote
						p++;
					}
					p=jsonbuf_skip_white(p);
					if(*p != ':')
					{
						// wrong token
						pt=token;
						continue;
					}
					p++;
					p=jsonbuf_skip_white(p);

					// token match
					if(token_count==n || n==-1)
					{
						have_token=true;
						res=p;
						switch(*p)
						{
							case '"' :
								typ=JSON_QUOTE;
								p++;
								do_val=true;
								break;
							case '\'' :
								typ=JSON_SQUOTE;
								p++;
								do_val=true;
								break;
							case '{' :
								p++;
								return p;
								break;
							case '[' :
								p++;
								if(n<0)
								{
									return p;
								}
								typ=JSON_ARRAY;
								break;
							default :
								typ=JSON_VAL;
								do_val=true;
								break;
						}
					}
					token_count++;
					continue;
				}
			}
			else
			{
				pt=token;
			}
		}
		else
		{
			if(do_val)
			{
				if(*p=='\\')
				{
					// escape 
					p++;
					if(*p)
					{
						if(pdst && rdst_len)
						{
							switch(*p)
							{
								case 'r':
									*pdst='\r';
									break;
								case 'n':
									*pdst='\n';
									break;
								case 't':
									*pdst='\t';
									break;
								case '\'':
									*pdst='\'';
									break;
								case '"':
									*pdst='"';
									break;
								case '\\':
									*pdst='\\';
									break;
								default :
									*pdst=*p;
									break;
							}
							pdst++;
							rdst_len--;
							// copy to dst
						}
						p++;
					}
					continue;
				}
			}

			switch(typ)
			{
				case JSON_ARRAY :
					if(array_count==n)
					{
						p=jsonbuf_skip_white(p);
						if(*p=='"')
						{
							typ=JSON_QUOTE;
							p++;
							do_val=true;
						}
						else if(*p=='\'')
						{
							typ=JSON_SQUOTE;
							p++;
							do_val=true;
						}
						else
						{
							typ=JSON_VAL;
							do_val=true;
						}
						continue;
					}
					else if(*p==']' || *p==',')
					{
						array_count++;
						p++;
						continue;
					}
					break;
				case JSON_QUOTE :
					if(*p=='"')
					{
						if(pdst&& rdst_len)
						{
							*pdst='\0';
							jsonbuf_print_elem(token,rdst);
							// copy to dst
						}
						return res;
					}
					break;
				case JSON_SQUOTE :
					if(*p=='\'')
					{
						if(pdst&& rdst_len)
						{
							*pdst='\0';
							// copy to dst
							jsonbuf_print_elem(token,rdst);
						}
						return res;
					}
					break;
				case JSON_VAL :
					if(*p<=' ' || *p==','|| *p=='}' || *p==']')
					{
						if(pdst&& rdst_len)
						{
							*pdst='\0';
							jsonbuf_print_elem(token,rdst);
						}
						return res;
					}
					break;
			}
			if(do_val)
			{
				if(pdst&& rdst_len)
				{
					*pdst++=*p;
					rdst_len--;
					// copy to dst
				}
			}
		}
		p++;
	}

	if(pdst&& rdst_len)
	{
		*pdst='\0';
	}

	// token not found
	return NULL; 
}

uint16_t jsonbuf_ident(uint8_t ident, char *buf)
{
	uint16_t len=0;

	while(ident)
	{
		buf[len]='\t';
		len++;
		ident--;
	}
	buf[len]='\0';
	return len;
}

uint16_t jsonbuf_str(char *val, char *buf)
{
	uint16_t len;

	len=strlen(val);
	strcpy(buf,val);
	return len;
}

uint16_t jsonbuf_char(uint8_t ident, char val, char *buf, bool have_komma)
{
	uint16_t len;

	len=jsonbuf_ident(ident, buf);
	buf[len]=val; len++;
	if(have_komma)
	{
		buf[len]=','; len++;
	}
	buf[len]='\n'; len++;
	buf[len]='\r'; len++;
	buf[len]='\0';
	return len;
}

uint16_t jsonbuf_obj_start(uint8_t ident, char *buf)
{
	return jsonbuf_char(ident, '{', buf,false);
}

uint16_t jsonbuf_obj_end(uint8_t ident, char *buf,bool have_komma)
{
	return jsonbuf_char(ident, '}', buf,have_komma);
}

uint16_t jsonbuf_array_end(uint8_t ident, char *buf,bool have_komma)
{
	return jsonbuf_char(ident, ']', buf,have_komma);
}

uint16_t jsonbuf_str(uint8_t ident, char *val, char *name, char *buf,bool have_komma,bool have_quote)
{
	uint16_t len;

	len=jsonbuf_ident(ident, buf);
	buf[len]='"'; len++;
	len+=jsonbuf_str(name, &buf[len]);
	buf[len]='"'; len++;
	buf[len]=':'; len++;
	if(have_quote)
	{
		buf[len]='"'; len++;
	}
	len+=jsonbuf_str(val, &buf[len]);
	if(have_quote)
	{
		buf[len]='"'; len++;
	}
	if(have_komma)
	{
		buf[len]=','; len++;
	}
	len+=jsonbuf_str((char *)"\n\r", &buf[len]);
	return len;
}

uint16_t jsonbuf_array_start(uint8_t ident, char *name, char *buf)
{
	uint16_t len;

	len=jsonbuf_ident(ident, buf);
	buf[len++]='"';
	len+=jsonbuf_str(name, buf+len);
	len+=jsonbuf_str((char *)"\":\n\r", buf+len);
	len+=jsonbuf_char(ident, '[', buf+len,false);
	return len;
}

uint16_t jsonbuf_float(uint8_t ident, float val, char *name, char *buf,bool have_komma)
{
	char out[12];
	dtostrf(val, 1, 4, out);

	return jsonbuf_str(ident, out, name, buf,have_komma,false);
}

uint16_t jsonbuf_int(uint8_t ident, int val, char *name, char *buf,bool have_komma)
{
	char out[12];
	itoa(val, out,10);

	return jsonbuf_str(ident, out, name, buf,have_komma,false);
}

#ifdef ttt
boolean save_pump(void)
{
	bool have_pump=false;
	uint16_t len;
	uint8_t n;
	File fi = SPIFFS.open(MFS_CONF_PUMPE, "w");
	if (!fi) 
	{
		Log.println(F("Failed to open pump-config file"));
		return false;
	}

	len=jsonbuf_obj_start(0, json_buf);
	len+=jsonbuf_array_start(1, (char *)"pump", &json_buf[len]);

	for(n=0;n<PUMP_MAX;n++)
	{
		if(pump_all[n].active&PUMP_ACT_IN_SYSTEM)
		{
			if(have_pump)
			{
				len+=jsonbuf_obj_end(2, &json_buf[len],true);
			}
			have_pump=true;
			len+=jsonbuf_obj_start(2, &json_buf[len]);
			len+=jsonbuf_int(3,n, (char *)"num", &json_buf[len],true);
			len+=jsonbuf_float(3,pump_all[n].ml, (char *)"ml", &json_buf[len],true);
			len+=jsonbuf_float(3,pump_all[n].retreat, (char *)"retreat", &json_buf[len],true);
			len+=jsonbuf_str(3,pump_all[n].name, (char *)"name", &json_buf[len],true,true);
			len+=jsonbuf_str(3,pump_all[n].typ, (char *)"typ", &json_buf[len],true,true);
			len+=jsonbuf_float(3,pump_all[n].pg, (char *)"PG", &json_buf[len],true);
			len+=jsonbuf_float(3,pump_all[n].vg, (char *)"VG", &json_buf[len],true);
			len+=jsonbuf_float(3,pump_all[n].h2o, (char *)"H2O", &json_buf[len],true);
			len+=jsonbuf_float(3,pump_all[n].nikotin, (char *)"NIKOTIN", &json_buf[len],true);
			len+=jsonbuf_float(3,pump_all[n].max, (char *)"max", &json_buf[len],false);
		}
	}
	if(have_pump)
	{
		len+=jsonbuf_obj_end(2, &json_buf[len],false);
	}
	len+=jsonbuf_array_end(1, &json_buf[len],false);
	len+=jsonbuf_obj_end(0, &json_buf[len],false);
	Log.print(json_buf);
	fi.print(json_buf);
	fi.close();
	return true;
}

boolean del_file(char *fname)
{
	return SPIFFS.remove(fname);
}

boolean save_rezept(char *fname)
{
	bool have_pump=false;
	uint16_t len;
	uint8_t n;

	Log.print(F("Save Rezept file: "));
	Log.println(fname);

	File fi = SPIFFS.open(fname, "w");
	if (!fi) 
	{
		Log.println(F("Failed to open rezept file"));
		return false;
	}

	len=jsonbuf_obj_start(0, json_buf);
	len+=jsonbuf_float(1,pump_rezept.vg, (char *)"vg", &json_buf[len],true);
	len+=jsonbuf_float(1,pump_rezept.pg, (char *)"pg", &json_buf[len],true);
	len+=jsonbuf_float(1,pump_rezept.h2o, (char *)"h2o", &json_buf[len],true);
	len+=jsonbuf_float(1,pump_rezept.nikotin, (char *)"nikotin", &json_buf[len],true);
	len+=jsonbuf_array_start(1, (char *)"aroma", &json_buf[len]);

	for(n=0;n<PUMP_MAX;n++)
	{
		if(pump_rezept.aroma[n].name[0]!='\0')
		{
			if(have_pump)
			{
				len+=jsonbuf_obj_end(2, &json_buf[len],true);
			}
			have_pump=true;
			len+=jsonbuf_obj_start(2, &json_buf[len]);
			len+=jsonbuf_str(3,pump_rezept.aroma[n].name, (char *)"name", &json_buf[len],true,true);
			len+=jsonbuf_float(3,pump_rezept.aroma[n].val, (char *)"volume", &json_buf[len],false);
		}
	}
	if(have_pump)
	{
		len+=jsonbuf_obj_end(2, &json_buf[len],false);
	}
	len+=jsonbuf_array_end(1, &json_buf[len],false);
	len+=jsonbuf_obj_end(0, &json_buf[len],false);
	Log.print(mfs_json_buf);
	fi.print(mfs_json_buf);
	fi.close();
	return true;
}

boolean load_rezept(char *fname)
{
	char *pt;
	char buf1[128];

	Log.print(F("Load Rezept file: "));
	Log.println(fname);

	File fi = SPIFFS.open(fname, "r");
	if (!fi) 
	{
		Log.println(F("Failed to open Rezept file"));
		return false;
	}

	size_t size = fi.size();
	if (size > (JSON_BSIZE-1)) 
	{
		fi.close();
		Log.println(F("Rezept file size is too large"));
		return false;
	}

	fi.readBytes(json_buf, size);
	json_buf[size]='\0';

	pt=json_parse(json_buf, (char *)"aroma", NULL, sizeof(buf1)-1, -1);
	if(!pt)
	{
		Log.println(F("Error parse"));
	}
	else
	{
		uint8_t i;
		char *ps;

		*pump_rezept.name=0;
		pump_rezept.nikotin=0.0;
		pump_rezept.pg=0.0;
		pump_rezept.vg=0.0;
		pump_rezept.h2o=0.0;

		for(i=0;i<PUMP_MAX;i++)
		{
			if(json_parse(pt, (char *)"name", buf1, sizeof(buf1)-1, i)) 
			{
				strncpy(pump_rezept.aroma[i].name,buf1,PUMP_NAME_MAX-1); pump_rezept.aroma[i].name[PUMP_NAME_MAX-1]='\0';
				if(json_parse(pt, (char *)"volume", buf1, sizeof(buf1)-1, i)) pump_rezept.aroma[i].val=atof(buf1); 
			}
			else
			{
				// this ist not set, zero it
				pump_rezept.aroma[i].name[0]='\0';
				pump_rezept.aroma[i].val=0.0;
			}
		}
		if(json_parse(json_buf, (char *)"h2o", buf1, sizeof(buf1)-1, -1)) pump_rezept.h2o=atof(buf1);
		if(json_parse(json_buf, (char *)"vg", buf1, sizeof(buf1)-1, -1)) pump_rezept.vg=atof(buf1);
		if(json_parse(json_buf, (char *)"pg", buf1, sizeof(buf1)-1, -1)) pump_rezept.pg=atof(buf1);
		if(json_parse(json_buf, (char *)"nikotin", buf1, sizeof(buf1)-1, -1)) pump_rezept.nikotin=atof(buf1);

		strncpy(buf1,fname,sizeof(buf1)-2);
		// cut at '.'
		ps=menu_cut_dir_line(buf1);

		strncpy(pump_rezept.name,ps,PUMP_NAME_MAX-1);
	}

	fi.close();
	return true;
}

boolean load_pump(void)
{
	char *pt;
	char buf1[128];

	Log.println(F("Load Pump-Config file"));

	File fi = SPIFFS.open(MFS_CONF_PUMPE, "r");
	if (!fi) 
	{
		Log.println(F("Failed to open pump-config file"));
		return false;
	}

	size_t size = fi.size();
	if (size > (JSON_BSIZE-1)) 
	{
		fi.close();
		Log.println(F("Pump-config file size is too large"));
		return false;
	}

	fi.readBytes(json_buf, size);
	json_buf[size]='\0';

	pt=json_parse(json_buf, (char *)"pump", NULL, sizeof(buf1)-1, -1);
	if(!pt)
	{
		Log.println(F("Error parse"));
	}
	else
	{
		uint8_t i;
		for(i=0;i<PUMP_MAX;i++)
		{
			if(json_parse(pt, (char *)"num", buf1, sizeof(buf1)-1, i))
			{
				uint8_t num=atoi(buf1);
				if(num>=PUMP_MAX) num=PUMP_MAX-1;
				if(json_parse(pt, (char *)"ml", buf1, sizeof(buf1)-1, i)) pump_all[num].ml=atof(buf1);
				if(json_parse(pt, (char *)"retreat", buf1, sizeof(buf1)-1, i)) pump_all[num].retreat=atof(buf1);
				if(json_parse(pt, (char *)"name", buf1, sizeof(buf1)-1, i)) { strncpy(pump_all[num].name,buf1,PUMP_NAME_MAX-1); pump_all[num].name[PUMP_NAME_MAX-1]='\0';}
				if(json_parse(pt, (char *)"typ", buf1, sizeof(buf1)-1, i)) { strncpy(pump_all[num].typ,buf1,PUMP_NAME_MAX-1); pump_all[num].typ[PUMP_NAME_MAX-1]='\0';}
				if(json_parse(pt, (char *)"PG", buf1, sizeof(buf1)-1, i)) pump_all[num].pg=atof(buf1);
				if(json_parse(pt, (char *)"VG", buf1, sizeof(buf1)-1, i)) pump_all[num].vg=atof(buf1);
				if(json_parse(pt, (char *)"H2O", buf1, sizeof(buf1)-1, i)) pump_all[num].h2o=atof(buf1);
				if(json_parse(pt, (char *)"NIKOTIN", buf1, sizeof(buf1)-1, i)) pump_all[num].nikotin=atof(buf1);
				if(json_parse(pt, (char *)"max", buf1, sizeof(buf1)-1, i)) pump_all[num].max=atof(buf1);
			}
		}
	}
	fi.close();
	return true;
}

bool loadConfig(void) 
{
	char buf1[128];

	Log.println(F("Load config file"));
	File fi = SPIFFS.open(MFS_CONF_CONFIG, "r");
	if (!fi) 
	{
		Log.println(F("Failed to open config file"));
		return false;
	}

	size_t size = fi.size();
	if (size > (MFS_JSON_BSIZE-1)) 
	{
		fi.close();
		Log.println(F("Config file size is too large"));
		return false;
	}

	fi.readBytes(json_buf, size);
	mfs_json_buf[size]='\0';

	if(json_parse(mfs_json_buf, (char *)"VG-mass", buf1, sizeof(buf1)-1, -1)) pump_vg_mass=atof(buf1);
	if(json_parse(mfs_json_buf, (char *)"PG-mass", buf1, sizeof(buf1)-1, -1)) pump_pg_mass=atof(buf1);
	if(json_parse(mfs_json_buf, (char *)"H2O-mass", buf1, sizeof(buf1)-1, -1)) pump_h2o_mass=atof(buf1);
	if(json_parse(mfs_json_buf, (char *)"ScaleCalib", buf1, sizeof(buf1)-1, -1)) scaleCalib=atof(buf1);
	if(json_parse(mfs_json_buf, (char *)"Methode", buf1, sizeof(buf1)-1, -1)) pump_methode=atoi(buf1);
	if(json_parse(mfs_json_buf, (char *)"Step", buf1, sizeof(buf1)-1, -1)) pump_step=atoi(buf1);
	if(json_parse(mfs_json_buf, (char *)"UPM_fast", buf1, sizeof(buf1)-1, -1)) pump_upm_fast=atoi(buf1);
	if(json_parse(mfs_json_buf, (char *)"UPM_normal", buf1, sizeof(buf1)-1, -1)) pump_upm_normal=atoi(buf1);
	if(json_parse(mfs_json_buf, (char *)"UPM_slow", buf1, sizeof(buf1)-1, -1)) pump_upm_slow=atoi(buf1);
	if(json_parse(mfs_json_buf, (char *)"UPM_xslow", buf1, sizeof(buf1)-1, -1)) pump_upm_xslow=atoi(buf1);
	if(json_parse(mfs_json_buf, (char *)"config_name", buf1, sizeof(buf1)-1, -1)) { strncpy(config_name,buf1,PUMP_NAME_MAX-1); config_name[PUMP_NAME_MAX-1]='\0';}
	fi.close();
	return true;
}

/**
   Save configuration data to a JSON file
   on SPIFFS
*/
void gen_saveConfig(void) 
{
	uint16_t len;

	len=jsonbuf_obj_start(0, json_buf);
	len+=jsonbuf_str(1,config_name, (char *)"config_name", &json_buf[len],true,true);
	len+=jsonbuf_float(1,pump_vg_mass, (char *)"VG-mass", &json_buf[len],true);
	len+=jsonbuf_float(1,pump_pg_mass, (char *)"PG-mass", &json_buf[len],true);
	len+=jsonbuf_float(1,pump_h2o_mass, (char *)"H2O-mass", &json_buf[len],true);
	len+=jsonbuf_float(1,scaleCalib, (char *)"ScaleCalib", &json_buf[len],true);
	len+=jsonbuf_int(1,pump_upm_fast, (char *)"UPM_fast", &json_buf[len],true);
	len+=jsonbuf_int(1,pump_upm_normal, (char *)"UPM_normal", &json_buf[len],true);
	len+=jsonbuf_int(1,pump_upm_slow, (char *)"UPM_slow", &json_buf[len],true);
	len+=jsonbuf_int(1,pump_upm_xslow, (char *)"UPM_xslow", &json_buf[len],true);
	len+=jsonbuf_int(1,pump_step, (char *)"Step", &json_buf[len],true);
	len+=jsonbuf_int(1,pump_methode, (char *)"Methode", &json_buf[len],false);
	len+=jsonbuf_obj_end(0, &json_buf[len],false);
}

bool saveConfig(char *p)
{
	File fi = SPIFFS.open(MFS_CONF_CONFIG, "w");
	if (!fi) 
	{
		Log.println(F("Failed to open Config file"));
		return false;
	}

	Log.print(p);
	fi.print(p);
	fi.close();
	return true;
}

bool saveConfig(void) 
{
	gen_saveConfig();
	return saveConfig(mfs_json_buf);
}

void resetsettings(WiFiManager& wifim)
{
	SPIFFS.format();
	wifim.resetSettings();
}

void init(void)
{
	//Load config upon start
	if (!SPIFFS.begin()) 
	{
		Log.println("Failed to mount file system");
		return;
	}
	loadConfig(); 
//	load_rezept((char *)"/rezepte/dampf.json");

}
#endif
